# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2019)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import textwrap as txtw
from pathlib import Path as path_t
from typing import Tuple

from numpy import ndarray as array_t

from im_tools_36.input import AvailableModules, ImageVolumeOrSequence

PATH_JR = "/home/eric/Data/Videos/people/jeremie-roux"


modules_details = AvailableModules()

print("--- MODULES")
for module in AvailableModules(as_modules=True):
    print(module)
print("--- MODULE DETAILS")
for pypi_name, import_name, version, function in modules_details:
    print(
        f"Pypi name: {pypi_name}\n"
        f"    Import name: {import_name}=={version}\n"
        f"    Function:    {function}"
    )


def CheckImage(path: path_t, expected_shape: Tuple[int | None, ...], /) -> None:
    """"""
    print(f"--- {path}: {expected_shape}")

    successes = []
    failures = []
    for module_ in ((None,),) + modules_details:
        ivs = ImageVolumeOrSequence(
            path,
            expected_shape=expected_shape,
            with_module=module_[0],
            should_print_module=module_[0] is None,
        )
        if isinstance(ivs, array_t):
            successes.append(f"{module_[0]}: Success w/ shape {ivs.shape}")
        else:
            exception = txtw.indent("\n".join(ivs), "    ")
            failures.append(f"!!! {module_[0]}: Failure\n{exception}")

    successes = txtw.indent("\n".join(successes), "    ")
    failures = txtw.indent("\n".join(failures), "    ")
    print(f"SUCCESSE(S) ---\n{successes}\nFAILURE(S) ---\n{failures}")


channels = ("YFP", "mCherry", "POL")
CheckImage(
    path_t(PATH_JR) / "p53" / "treat01_10_R3D.dv", (None, None, channels.__len__(), 433)
)

channels = ("CFP", "YFP", "POL")
CheckImage(path_t(PATH_JR) / "TREAT02_13_R3D.dv", (None, None, channels.__len__(), 241))

channels = ("CFP", "YFP", "POL")
CheckImage(
    path_t(PATH_JR)
    / "HeLaJR2_treat-01-Scene-16-P16-D03"
    / "20231123_LCM_HeLaJR2_CGprep2_treat-01-Scene-16-P16-D03.czi",
    (None, None, channels.__len__(), 358),
)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2019)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence, Union

import imageio as imio
import numpy as nmpy
import tifffile as tiff

from im_tools_36.path import path_h


array_t = nmpy.ndarray


def SaveImage8(image: array_t, path: path_h, /, *, w_scaling: bool = True) -> None:
    """
    8: For 8 bits
    """
    image = image.astype(nmpy.float64, copy=False)
    if w_scaling:
        img_min = nmpy.min(image)
        img_max = nmpy.max(image)
        if img_max > img_min:
            image = (255.0 / (img_max - img_min)) * (image - img_min)
        else:
            # Avoid -= in case it really is inplace (since copy=False was used above)
            image = image - img_min
    image = nmpy.around(image).astype(nmpy.uint8)

    imio.imwrite(path, image)


def SaveSequenceAsTIFF(
    sequence: Union[array_t, Sequence[array_t]],
    path: path_h,
    /,
    *,
    has_channels: bool = None,
    channels: Sequence[str] = None,
    logger=None,
) -> None:
    """
    This function is meant to deal with 2-D images or sequences of 2-D images, possibly multi-channel. It does not deal
    with 3-D images.

    If "sequence" is a Numpy array, it must be with dimensions: XY, XYC, XYT, or XYCT. If it is a sequence of arrays,
    their stacking along a new, final axis must result in an array with the previously mentioned dimensions.

    sequence: or image actually
    has_channels: to be passed only for XYC and XYT. An exception is raised otherwise.
    """
    if isinstance(sequence, array_t):
        array = sequence
    else:
        array = nmpy.stack(sequence, axis=sequence[0].ndim)

    if (has_channels is not None) and (array.ndim != 3):
        raise ValueError(
            'Parameter "has_channels" must be passed with XYC and XYT inputs only'
        )

    if logger is None:
        EmitWarning = print
    else:
        EmitWarning = logger.warning

    comments = []

    if nmpy.issubsctype(array.dtype, bool):
        array = array.astype(nmpy.uint8)
        array[array > 0] = 255
        comments.append(f"Original type: {array.dtype.name}\nFalse -> 0\nTrue -> 255")

    if array.ndim == 2:
        pages = 1
        shape = array.shape
        axes = "XY"
        planar_config = "separate"
    elif array.ndim == 3:
        array = nmpy.moveaxis(array, (0, 1, 2), (1, 2, 0))
        if has_channels:
            pages = 1
            shape = array.shape
            axes = "CXY"
            planar_config = "separate"
        else:
            pages = array.shape[0]
            shape = array.shape[1:]
            axes = "TXY"
            planar_config = None
    elif array.ndim == 4:
        if not nmpy.issubdtype(array.dtype, nmpy.uint8):
            for c_idx in range(array.shape[2]):
                channel = array[..., c_idx, :]
                minimum, maximum = nmpy.amin(channel), nmpy.amax(channel)
                if maximum == minimum:
                    normalized = nmpy.zeros(channel.shape, dtype=nmpy.uint8)
                else:
                    normalized = nmpy.around(
                        255.0 * ((channel - minimum) / (maximum - minimum))
                    ).astype(nmpy.uint8)
                array[..., c_idx, :] = normalized
                comments.append(f"Channel {c_idx}: min={minimum}, max={maximum}")
            comments.append(f"Original type: {array.dtype.name}")
            EmitWarning(f"Downtyping from {array.dtype.name} to uint8")
        array = nmpy.moveaxis(array, (0, 1, 2, 3), (2, 3, 1, 0))
        pages = array.shape[0]
        shape = array.shape[1:]
        axes = "TCXY"
        planar_config = "separate"
    else:
        raise ValueError(f"{array.ndim}: Unhandled image dimension")

    meta_data = {
        "pages": pages,
        "shape": shape,
        "axes": axes,
        "ImageDescription": "\n".join(comments),
    }
    if channels is not None:
        meta_data.update({"Channel": {"Name": channels}})

    tiff.imwrite(
        str(path),
        data=array,
        photometric="minisblack",
        compression="deflate",
        planarconfig=planar_config,
        metadata=meta_data,
    )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2019)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Formats:
    DeltaVision
        dv, r3d, rcpnl
        https://docs.openmicroscopy.org/bio-formats/6.9.1/formats/deltavision.html

imageio:
    See https://imageio.readthedocs.io/en/stable/reference/index.html#plugins-backend-libraries

Using VTK seems too complicated. See
https://stackoverflow.com/questions/25230541/how-to-convert-a-vtkimage-into-a-numpy-array
for a 3-D example.
"""
import importlib as mprt
import itertools as ittl
import sys as sstm
from functools import partial as partial_t
from types import ModuleType as module_t
from typing import Callable, Sequence, Tuple

import numpy as nmpy

from im_tools_36.path import path_h, path_t

array_t = nmpy.ndarray


def _WithAICS(aics: module_t, path: str, /) -> array_t:
    """
    Arrangement: TCZYX
    """
    return aics.AICSImage(path).data


def _WithCziTools(czit: module_t, path: str, /) -> array_t:
    """"""
    output, *_ = czit.read_6darray(path)
    return output


def _WithMRC(mrci: module_t, path: str, /) -> array_t:
    """
    If output.ndim == 5, probably time x channel x Z x Y x X, while sequences are:
        time x channel   x (Z=1 x)             Y x X.
    So one gets:
        time x channel=1 x Z=actual channels x Y x X. Then, use: output[:, 0, :, :]
    numpy.array: Because the returned value seems to be a read-only memory map
    """
    return nmpy.array(mrci.imread(path))


def _WithNibabel(nbbl: module_t, path: str, /) -> array_t:
    """"""
    image = nbbl.load(path)

    return image.get_fdata()


def _WithPillow(pllw: module_t, path: str, /) -> array_t:
    """"""
    image_t = getattr(pllw, "Image")
    with image_t.open(path) as image:
        image.load()

    return nmpy.asarray(image)


# Key:   Pypi-name of toplevel package on Pypi.
# Value: Name of (sub)-package to be imported ("": Same as key).
IMPORT_FOR_MODULE = {
    "aicsimageio": ("", _WithAICS),
    "czifile": ("", "imread"),
    "czitools": ("czitools.read_tools", _WithCziTools),
    "imageio": ("imageio.v3", "imread"),
    "itk": ("", "imread"),
    "mrc": ("", _WithMRC),
    "nibabel": ("", _WithNibabel),
    "opencv": ("cv2", "imread"),
    "pillow": ("PIL", _WithPillow),
    "scikit-image": ("skimage.io", "imread"),
    "tifffile": ("", "imread"),
}
# Aliases
IMPORT_FOR_MODULE["skimage"] = IMPORT_FOR_MODULE["scikit-image"]

# Modules must be ordered from general to specific.
# aicsimageio is placed last for CZI format because czifile and czitools add a spurious
# row or column.
ORDERED_MODULES_PER_TYPE = {
    (): ("imageio", "itk", "pillow", "opencv", "scikit-image", "nibabel"),
    ("tif", "tiff"): ("aicsimageio", "tifffile"),
    ("dv",): ("aicsimageio", "mrc"),
    ("mrc",): ("mrc",),
    ("czi", "lif", "nd2"): ("czifile", "czitools", "aicsimageio"),
}


def _Module(name: str, /) -> module_t | None:
    """"""
    if name in sstm.modules:
        return sstm.modules[name]

    # https://docs.python.org/3/library/importlib.html#checking-if-a-module-can-be-imported
    if (spec := mprt.util.find_spec(name)) is None:
        return None

    output = mprt.util.module_from_spec(spec)
    sstm.modules[name] = output
    spec.loader.exec_module(output)

    return output


def _ImportDictionaries(
    import_for_module: dict[
        str, tuple[str | module_t, str | Callable[[module_t, path_h], array_t]]
    ]
) -> tuple[dict[str, module_t], dict[str, Callable[[path_h], array_t]]]:
    """"""
    module_with_name = {}
    function_of_module = {}

    for pypi_name, (import_name, Function) in import_for_module.items():
        if import_name == "":
            import_name = pypi_name
        module = _Module(import_name)
        if module is None:
            continue

        if isinstance(Function, str):
            Function = getattr(module, Function)
        else:
            # lambda functions are not adapted here (to be investigated some day).
            Function = partial_t(Function, module)

        module_with_name[pypi_name] = module
        function_of_module[pypi_name] = Function

    return module_with_name, function_of_module


MODULE_WITH_NAME, FUNCTION_OF_MODULE = _ImportDictionaries(IMPORT_FOR_MODULE)


def ImageVolumeOrSequence(
    path: path_h,
    /,
    *,
    should_squeeze: bool = True,
    expected_dim: int = None,
    expected_shape: Sequence[int | None] = None,
    with_module: str = None,
    should_print_module: bool = False,
) -> array_t | Tuple[str, ...]:
    """"""
    # Potential outputs
    image = None
    issues = []

    if isinstance(path, str):
        path_str = path
        path_lib = path_t(path)
    else:
        path_str = str(path)
        path_lib = path
    if (expected_dim is None) and (expected_shape is not None):
        expected_dim = expected_shape.__len__()

    if with_module is None:
        reading_functions = list(
            FUNCTION_OF_MODULE.get(_elm) for _elm in ORDERED_MODULES_PER_TYPE[()]
        )

        img_format = path_lib.suffix[1:].lower()
        for formats, modules in ORDERED_MODULES_PER_TYPE.items():
            if img_format in formats:
                reading_functions.extend(
                    FUNCTION_OF_MODULE.get(_elm) for _elm in modules
                )

        reading_functions = tuple(
            _elm for _elm in reading_functions if _elm is not None
        )
        if reading_functions.__len__() == 0:
            return (f"No modules available for image format {img_format}.",)
    elif with_module in FUNCTION_OF_MODULE:
        reading_functions = [FUNCTION_OF_MODULE[with_module]]
    else:
        return (
            f"{with_module}: Invalid module. "
            f"Expected={str(tuple(FUNCTION_OF_MODULE.keys()))[1:-1]}",
        )

    failure = True
    for Read in reversed(reading_functions):
        try:
            image = Read(path_str)
            # A module might return None, for example, in case of a failure instead of
            # raising an exception. Hence the test below.
            if isinstance(image, array_t):
                failure = False
                if should_print_module:
                    if isinstance(Read, partial_t):
                        Read = Read.func
                    print(
                        f'{path_str}: Read with function "{Read.__module__}.{Read.__name__}".'
                    )
                break
        except Exception as exception:
            if isinstance(Read, partial_t):
                Read = Read.func
            issues.append(
                f'Cannot open image with function "{Read.__module__}.{Read.__name__}".\n'
                f"Error:\n{exception}"
            )

    if failure:
        if issues.__len__() > 0:
            return tuple(issues)
        return ("Silent Exception",)

    if should_squeeze:
        image = nmpy.squeeze(image)

    if (expected_dim is not None) and (image.ndim != expected_dim):
        return (
            f"{image.ndim}: Invalid dimension (shape={image.shape}). "
            f"Expected={expected_dim}",
        )
    if expected_shape is None:
        return image

    shape = image.shape
    if _ShapeMatches(shape, expected_shape):
        return image

    shape_as_array = nmpy.array(shape)
    for order in ittl.permutations(range(image.ndim)):
        if _ShapeMatches(shape_as_array[nmpy.array(order)], expected_shape):
            return nmpy.moveaxis(image, order, range(image.ndim))

    return (
        f"{shape}: Invalid shape. "
        f"Expected={tuple(expected_shape)}, or a permutation of it.",
    )


def _ShapeMatches(
    actual: Tuple[int, ...] | array_t, expected: Sequence[int | None], /
) -> bool:
    """"""
    return all((_ctl == _xpt) or (_xpt is None) for _ctl, _xpt in zip(actual, expected))


def AvailableModules(
    *, as_modules: bool = False
) -> Tuple[module_t, ...] | Tuple[Tuple[str, str, str, str], ...]:
    """
    Tuple[str, str, str, str]:
        name for "with_module" parameter of ImageVolumeOrSequence,
        module Python name,
        module version,
        reading function name
    """
    if as_modules:
        return tuple(MODULE_WITH_NAME.values())

    output = []
    for pypi_name, module in MODULE_WITH_NAME.items():
        import_name = getattr(module, "__name__", "???")
        if "." in import_name:
            parent = import_name.split(sep=".", maxsplit=1)[0]
            version = getattr(sstm.modules[parent], "__version__", "???")
        else:
            version = getattr(module, "__version__", "???")
        function = FUNCTION_OF_MODULE[pypi_name]
        if isinstance(function, partial_t):
            function = function.func
        output.append((pypi_name, import_name, version, function.__name__))

    return tuple(output)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2019)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence

import numpy as nmpy
from skimage import transform as sitf


array_t = nmpy.ndarray


def Mosaic(imgs: Sequence[array_t], /, *, resize_factor: float = 1.0) -> array_t:
    #
    n_images = imgs.__len__()
    n_cols = int(nmpy.ceil(nmpy.sqrt(n_images)))
    n_rows, remaining = divmod(n_images, n_cols)
    if remaining > 0:
        n_rows += 1

    vignette_lengths = tuple(
        int(nmpy.around(resize_factor * length).item()) for length in imgs[0].shape
    )

    output = nmpy.empty(
        (n_rows * vignette_lengths[0], n_cols * vignette_lengths[1]), dtype=nmpy.uint64
    )

    row = 0
    col = 0
    row_slice = None
    for img in imgs:
        row_slice = slice(row * vignette_lengths[0], (row + 1) * vignette_lengths[0])
        col_slice = slice(col * vignette_lengths[1], (col + 1) * vignette_lengths[1])
        output[row_slice, col_slice] = sitf.resize(img, vignette_lengths)

        col += 1
        if col >= n_cols:
            col = 0
            row += 1
    col -= 1  # Unroll last unused col index

    if col < n_cols - 1:
        diagonal = nmpy.diag(vignette_lengths[0] * (1,)) > 0

        cross_value = nmpy.median(output)
        cross_img = nmpy.zeros(vignette_lengths)
        cross_img[diagonal] = cross_value
        cross_img[nmpy.rot90(diagonal)] = cross_value

        while col < n_cols - 1:
            col += 1
            col_slice = slice(
                col * vignette_lengths[1], (col + 1) * vignette_lengths[1]
            )
            output[row_slice, col_slice] = cross_img

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2019)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Tuple

import numpy as nmpy
from im_tools_36.type.image import intensity_h, number_h
from scipy import ndimage as spim
from skimage import color as sicl


array_t = nmpy.ndarray


def ValueScaledVersion(
    img: array_t, /, scaling: Tuple[intensity_h, intensity_h] = (-1.0, 1.0)
) -> array_t:
    """
    Value-scaled: as opposed to geometrically scaled.
    Default scaling is between [-1,+1], where the number representation is the most detailed
    TODO: check skimage.exposure.rescale_intensity
    """
    min_value, max_value, *_ = spim.extrema(img)
    factor = (scaling[1] - scaling[0]) / (max_value - min_value)
    output = factor * (img.astype(nmpy.float64, copy=False) - min_value) + scaling[0]

    return output


def UInt8RGBVersion(img: array_t, /, *, normalization: bool = True) -> array_t:
    """"""
    if normalization:
        normalized_img = ValueScaledVersion(img, scaling=(0, 255))
    else:
        normalized_img = img
    rounded_img = nmpy.around(normalized_img).astype(nmpy.uint8)

    if rounded_img.ndim == 2:
        output = nmpy.repeat(rounded_img[:, :, nmpy.newaxis], 3, axis=2)
    elif rounded_img.ndim == 3:
        if rounded_img.shape[2] == 1:
            output = nmpy.repeat(rounded_img, 3, axis=2)
        elif rounded_img.shape[2] == 3:
            output = rounded_img
        else:
            raise ValueError(
                f"{rounded_img.shape[2]}: Invalid number of planes; Must be 1 or 3"
            )
    else:
        raise ValueError(f"{rounded_img.ndim}: Invalid dimension; Must be 2 or 3")

    return output


def GloballyRelightedVersion(
    img: array_t, /, amplitude: intensity_h | number_h, *, mode: str = "+"
) -> array_t:
    """
    img: must have a uint8 dtype
    amplitude: in [0, 100] in additive mode, and in [0, infinity[ for multiplicative mode
    """
    relighted = sicl.rgb2lab(img)
    lightness = relighted[:, :, 0]
    if mode == "+":
        lightness += amplitude
    elif mode == "*":
        lightness *= amplitude
    else:
        raise ValueError(f"{mode}: Invalid mode")
    lightness[lightness < 0.0] = 0.0
    lightness[lightness > 100.0] = 100.0

    output = sicl.lab2rgb(relighted)
    output = UInt8RGBVersion(output)

    return output



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2020)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Final, Sequence

import numpy as nmpy
from im_tools_36.type.image import pix_value_h


array_t: Final = nmpy.ndarray


# TODO: deal with rgba images everywhere, maybe


def WithSuperimposedGrid(
    img: array_t,
    /,
    *,
    color: pix_value_h = 0,
    row_step: int = 4,
    col_step: int = 4,
    thickness: int = 1,
) -> array_t:
    """"""
    output = img.copy()

    if img.ndim == 2:
        if isinstance(color, Sequence):
            raise ValueError(
                f"Color must be a number, not a sequence, when image is two-dimensional"
            )
    elif img.ndim == 3:
        if img.shape[2] not in (3, 4):
            raise ValueError(
                f"{img.shape[2]}: Invalid length of 3rd image dimension; Must be 3 (RGB) or 4 (RGBA)"
            )
        if isinstance(color, Sequence):
            if color.__len__() != img.shape[2]:
                raise ValueError(
                    f"{color.__len__()}: Invalid number of color components; "
                    f"Must be {img.shape[2]} like the 3rd image dimension"
                )
        else:
            color = img.shape[2] * (color,)
    else:
        raise ValueError(
            f"{img.ndim}: Invalid number of image dimensions; Must be 2 (Grayscale) or 3 (RGB or RGBA)"
        )

    if isinstance(color, Sequence):
        color = nmpy.reshape(color, (1, 1, -1))

    for row in range(0, img.shape[0], row_step):
        output[row : (row + thickness), ...] = color
    for col in range(0, img.shape[1], col_step):
        output[:, col : (col + thickness), ...] = color

    return output
